{# Import the theme's layout. #}
{% extends "!layout.html" %}

{% block htmltitle %}
    {% if title == '' or title == 'Home' %}
        <title>{{ docstitle|striptags|e }}</title>
    {% else %}
        <title>{{ title|striptags|e }} | {{ docstitle|striptags|e }}</title>
    {% endif %}
{% endblock %}
